import logging
from typing import List, Optional, Tuple

import pandas

from hts._t import NAryTreeT
from hts.core.exceptions import InvalidArgumentException

logger = logging.getLogger(__name__)


def make_iterable(nodes: NAryTreeT, prop="key"):
    if not prop:
        it = [nodes] + nodes.traversal_level()
    else:
        it = [getattr(nodes, prop)] + [
            getattr(k, prop) for k in nodes.traversal_level()
        ]
    return it


def fetch_cols(exogenous, name) -> Tuple[List[str], Optional[List[str]]]:
    if not exogenous:
        return [name], None
    exog = exogenous.get(name, None)
    cols = [name] + exog if exog else [name]
    return cols, exog


def hexify(df, lat_col, lon_col, levels=(6, 8)) -> Optional[pandas.DataFrame]:
    try:
        from h3 import h3
    except ImportError:  # pragma: no cover
        logger.error(
            "h3-py must be installed for geo hashing capabilities. Exiting."
            "Install it with: pip install scikit-hts[geo]"
        )
        return
    for r in range(levels[0], levels[1] + 1):
        df[f"hex_index_{r}"] = df.apply(
            lambda x: h3.geo_to_h3(x[lat_col], x[lon_col], r), 1
        )
    return df


def resample_count(df: pandas.DataFrame, freq: str, colname: str) -> pandas.DataFrame:
    _df = pandas.DataFrame({colname: df.index}, index=df.index)
    return _df.resample(freq).agg("count")


def groupify(
    root_node, df, freq="1H", nodes=None, min_count=0.1, total=None
) -> NAryTreeT:
    """

    Parameters
    ----------
    fillna
    root_node : NAryTreeT
    df : pandas.DataFrame
    freq : str
        resample frequency
    nodes : tuple
        Hierarchy node
    min_count : int or float
        Minimum number of observations for a node to be used. If float, it represents a
        fraction of values that need to be non NaNs and must be between 0 and 1
    total : pandas.DataFrame:
        Name of root node

    Returns
    -------

    """

    child_group = nodes[0]  # city
    children = df[child_group].unique()  # berlin, munich, ...

    # add first level children
    for child in children:
        sub_df = df[df[child_group] == child]
        resampled = resample_count(sub_df, freq, child)
        root_node.add_child(key=child, item=resampled, exogenous=None)

    # add the rest
    for node in nodes[1:]:
        parent_group = child_group
        child_group = node  # hex_index_6
        children = df[child_group].unique()  # abcccc, abccf, ...

        for child in children:
            sub_df = df[df[child_group] == child]
            if isinstance(min_count, float):
                allowance = len(total) * min_count
            elif isinstance(min_count, int):
                allowance = min_count
            else:
                raise InvalidArgumentException(
                    "min_count must be either float or integer"
                )
            if len(sub_df) < allowance:
                continue
            parent_name = sub_df[parent_group].value_counts().index[0]
            resampled = resample_count(sub_df, freq, child)
            for c in root_node.traversal_level():
                if c.key == parent_name:
                    c.add_child(key=child, item=resampled, exogenous=None)
    return root_node



import weakref
from collections import deque
from itertools import chain
from typing import List, Optional, Tuple, Union

import pandas

from hts._t import ExogT, NAryTreeT, NodesT
from hts.hierarchy.utils import (
    fetch_cols,
    groupify,
    hexify,
    make_iterable,
    resample_count,
)
from hts.viz.geo import HierarchyVisualizer


class HierarchyTree(NAryTreeT):
    """
    A generic N-ary tree implementations, that uses a list to store
    it's children.
    """

    @classmethod
    def from_geo_events(
        cls,
        df: pandas.DataFrame,
        lat_col: str,
        lon_col: str,
        nodes: Tuple,
        levels: Tuple[int, int] = (6, 7),
        resample_freq: str = "1H",
        min_count: Union[float, int] = 0.2,
        root_name: str = "total",
        fillna: bool = False,
    ):
        """

        Parameters
        ----------
        df : pandas.DataFrame
        lat_col : str
            Column where the latitude coordinates can be found
        lon_col : str
            Column where the longitude coordinates can be found
        nodes : str

        levels :
        resample_freq
        min_count
        root_name
        fillna

        Returns
        -------
        HierarchyTree
        """

        hexified = hexify(df, lat_col, lon_col, levels=levels)
        total = resample_count(hexified, resample_freq, root_name)
        hierarchy = cls(key=root_name, item=total)
        grouped = groupify(
            hierarchy,
            df=hexified,
            nodes=nodes,
            freq=resample_freq,
            min_count=min_count,
            total=total,
        )
        # TODO: more flexible strategy
        if fillna:
            df = grouped.to_pandas()
            df = df.fillna(method="ffill").dropna()
            for node in make_iterable(grouped, prop=None):
                repl = df[[node.key]]
                node.item = repl
        return grouped

    @classmethod
    def from_nodes(
        cls,
        nodes: NodesT,
        df: pandas.DataFrame,
        exogenous: ExogT = None,
        root: Union[str, "HierarchyTree"] = "total",
        top: "HierarchyTree" = None,
        stack: List = None,
    ):
        """
        Standard method for creating a hierarchy from nodes and a dataframe containing as columns those nodes.
        The nodes are represented as a dictionary containing as keys the nodes, and as values list of edges.
        See the examples for usage. The total column must be named total and not something else.

        Parameters
        ----------
        nodes : NodesT
            Nodes definition. See ``Examples``.
        df : pandas.DataFrame
            The actual data containing the nodes
        exogenous : ExogT
            The nodes representing the exogenous variables
        root : Union[str, HierarchyTree]
            The name of the root node
        top : HierarchyTree
            Not to be used for initialisation, only in recursive calls
        stack : list
            Not to be used for initialisation, only in recursive calls

        Returns
        -------
        hierarchy : HierarchyTree
            The hierarchy tree representation of your data

        Examples
        --------
        In this example we will create a tree from some multivariate data

        >>> from hts.utilities.load_data import load_mobility_data
        >>> from hts.hierarchy import HierarchyTree

        >>> hmv = load_mobility_data()
        >>> hmv.head()
                    WF-01  CH-07  BT-01  CBD-13  SLU-15  CH-02  CH-08  SLU-01  BT-03  CH-05  SLU-19  SLU-07  SLU-02  CH-01  total   CH  SLU  BT  OTHER  temp  precipitation
        starttime
        2014-10-13     16     14     20      16      20     42     24      24     12     22      14       2       8      6    240  108   68  32     32  62.0           0.00
        2014-10-14     22     28     28      38      36     36     42      40     14     26      18      32      16     18    394  150  142  42     60  59.0           0.11
        2014-10-15     10     14      8      20      18     38     16      28     18     10       0      24      10     16    230   94   80  26     30  58.0           0.45
        2014-10-16     22     18     24      44      44     40     24      20     22     18       8      26      14     14    338  114  112  46     66  61.0           0.00
        2014-10-17      8     12     16      20      18     22     32      12      8     28      10      30       8     10    234  104   78  24     28  60.0           0.14

        >>> hier = {
                'total': ['CH', 'SLU', 'BT', 'OTHER'],
                'CH': ['CH-07', 'CH-02', 'CH-08', 'CH-05', 'CH-01'],
                'SLU': ['SLU-15', 'SLU-01', 'SLU-19', 'SLU-07', 'SLU-02'],
                'BT': ['BT-01', 'BT-03'],
                'OTHER': ['WF-01', 'CBD-13']
            }
        >>> exogenous = {k: ['precipitation', 'temp'] for k in hmv.columns if k not in ['precipitation', 'temp']}
        >>> ht = HierarchyTree.from_nodes(hier, hmv, exogenous=exogenous)
        >>> print(ht)
        - total
           |- CH
           |  |- CH-07
           |  |- CH-02
           |  |- CH-08
           |  |- CH-05
           |  - CH-01
           |- SLU
           |  |- SLU-15
           |  |- SLU-01
           |  |- SLU-19
           |  |- SLU-07
           |  - SLU-02
           |- BT
           |  |- BT-01
           |  - BT-03
           - OTHER
              |- WF-01
              - CBD-13

        """

        if stack is None:
            stack = []
            cols, ex = fetch_cols(exogenous, root)
            root = HierarchyTree(
                key=root, item=df[cols], exogenous=ex
            )  # root node is created
            top = root  # it is stored in b variable
        x = root.key  # root.val = 2 for the first time
        if len(nodes[x]) > 0:  # check if there are children of the node exists or not
            for i in range(len(nodes[x])):  # iterate through each child
                key = nodes[x][i]
                cols, ex = fetch_cols(exogenous, key)
                y = HierarchyTree(
                    key=key, item=df[cols], exogenous=ex
                )  # create Node for every child
                root.children.append(y)  # append the child_node to its parent_node
                stack.append(y)  # store that child_node in stack
                if (
                    y.key not in nodes.keys()
                ):  # if the child_node_val = 0 that is the parent = leaf
                    stack.pop()  # pop the 0 value from the stack
            if len(stack):
                if len(stack) >= 2:  # bottom-to-top
                    tmp_root = stack.pop(0)
                else:
                    tmp_root = stack.pop()
                cls.from_nodes(
                    nodes=nodes,
                    df=df,
                    exogenous=exogenous,
                    root=tmp_root,
                    top=top,
                    stack=stack,
                )  # pass tmp_root to the function as root
        return top

    def __init__(
        self,
        key: str = None,
        item: Union[pandas.Series, pandas.DataFrame] = None,
        exogenous: List[str] = None,
        children: List[NAryTreeT] = None,
        parent: NAryTreeT = None,
    ):

        self.key = key
        self.item = item
        if exogenous:
            self.exogenous = exogenous
        else:
            self.exogenous = []
        self.children = children or []
        self._parent = weakref.ref(parent) if parent else None
        self.visualizer = HierarchyVisualizer(self)

    def get_node(self, key: str) -> Optional[NAryTreeT]:
        """
        Get a node given its key

        Parameters
        ----------
        key: str
            The key of the node of interest
        Returns
        -------
        node : HierarchyTree
            The node of interest

        """
        for node in self.traversal_level():
            if node.key == key:
                return node
        return None

    def traversal_level(self) -> List[NAryTreeT]:
        """
        Level order traversal of the tree

        Returns
        -------
        list of nodes
        """

        res = []
        q = deque([(self, 0)])
        while q:
            n, li = q.popleft()
            if len(res) < li + 1:
                res.append([])
            for i in n.children:
                q.append((i, li + 1))
            res[li].extend(n.children)
        return list(chain.from_iterable(res[:-1]))

    def num_nodes(self) -> int:
        """
        Return the of nodes in the tree

        Returns
        -------
        num nodes : int
        """

        return sum(chain.from_iterable(self.level_order_traversal()))

    def is_leaf(self):
        """
        Check if node is a leaf Node

        Returns
        -------
        bool
            True or False
        """

        return len(self.children) == 0

    def value_at_height(self, level: int) -> List:
        if level == 0:
            return [1]
        return self.level_order_traversal()[level - 1]

    def sum_at_height(self, level) -> int:
        return sum(self.value_at_height(level))

    def get_height(self) -> int:
        return len(self.level_order_traversal())

    def get_node_height(self, key: str) -> int:
        for node in self.traversal_level():
            if node.key == key:
                return node.get_height()
        return -1

    def level_order_traversal(self: NAryTreeT) -> List[List[int]]:
        """
        Iterate through the tree in level order, getting the number of children for
        each node

        Returns
        -------
        list[list[int]]
        """

        res = []
        q = deque([(self, 0)])
        while q:
            n, li = q.popleft()
            if len(res) < li + 1:
                res.append([])
            for i in n.children:
                q.append((i, li + 1))
            res[li].append(len(n.children))
        return res[:-1]

    def get_level_order_labels(self: NAryTreeT) -> List[List[str]]:
        """
        Get the associated node labels from the NAryTreeT level_order_traversal().

        Parameters
        ----------
        self: NAryTreeT
            Tree being searched.

        Returns
        -------
        List[List[str]]
            Node labels corresponding to level order traversal.
        """
        labels = []
        q = deque([(self, 0)])
        while q:
            n, li = q.popleft()
            if len(labels) < li + 1:
                labels.append([])
            for i in n.children:
                q.append((i, li + 1))
            labels[li].append(n.key)
        return labels

    def add_child(self, key=None, item=None, exogenous=None) -> NAryTreeT:
        child = HierarchyTree(key=key, item=item, exogenous=exogenous, parent=self)
        self.children.append(child)
        return child

    def leaf_sum(self) -> int:
        return sum(self.level_order_traversal()[-1])

    def to_pandas(self) -> pandas.DataFrame:
        """
        Transforms the hierarchy into a pandas.DataFrame
        Returns
        -------
        df : pandas.DataFrame
            Dataframe representation of the tree
        """

        df = pandas.concat([self.item] + [c.item for c in self.traversal_level()], 1)
        df.index.name = "ds"
        return df

    def get_series(self) -> pandas.Series:
        return self.item[self.key]



import numpy
import pandas
from fbprophet import Prophet
from pmdarima import AutoARIMA

from hts.model import AutoArimaModel, FBProphetModel, HoltWintersModel, SarimaxModel
from hts.model.base import TimeSeriesModel


def test_instantiate_fb_model_uv(uv_tree):
    fb = FBProphetModel(node=uv_tree)
    assert isinstance(fb, TimeSeriesModel)
    fb = FBProphetModel(node=uv_tree, capacity_max=1)
    assert isinstance(fb, TimeSeriesModel)
    fb = FBProphetModel(node=uv_tree, capacity_min=1)
    assert isinstance(fb, TimeSeriesModel)


def test_instantiate_fb_model_mv(mv_tree):
    fb = FBProphetModel(node=mv_tree)
    assert isinstance(fb, TimeSeriesModel)


def test_fit_predict_fb_model_uv(uv_tree):
    fb = FBProphetModel(node=uv_tree)
    fb.fit()
    assert isinstance(fb.model, Prophet)
    fb.predict(uv_tree)
    assert isinstance(fb.forecast, pandas.DataFrame)
    assert isinstance(fb.residual, numpy.ndarray)
    assert isinstance(fb.mse, float)


def test_fit_predict_ar_model_uv(uv_tree):
    ar = AutoArimaModel(node=uv_tree,)
    ar.fit(max_iter=1)
    assert isinstance(ar.model, AutoARIMA)
    ar.predict(uv_tree)
    assert isinstance(ar.forecast, pandas.DataFrame)
    assert isinstance(ar.residual, numpy.ndarray)
    assert isinstance(ar.mse, float)


def test_fit_predict_sarimax_model_uv(uv_tree):
    sar = SarimaxModel(node=uv_tree, max_iter=1,)
    fitted_sar = sar.fit()
    assert isinstance(fitted_sar, SarimaxModel)
    sar.predict(uv_tree)
    assert isinstance(sar.forecast, pandas.DataFrame)
    assert isinstance(sar.residual, numpy.ndarray)
    assert isinstance(sar.mse, float)


def test_fit_predict_hw_model_uv(uv_tree):
    hw = HoltWintersModel(node=uv_tree,)
    fitted_hw = hw.fit()
    assert isinstance(fitted_hw, HoltWintersModel)
    hw.predict(uv_tree)
    assert isinstance(hw.forecast, pandas.DataFrame)
    assert isinstance(hw.residual, numpy.ndarray)
    assert isinstance(hw.mse, float)



from typing import Dict, Optional

import numpy
import pandas

from hts._t import MethodT, NAryTreeT, TransformT
from hts.functions import to_sum_mat
from hts.revision import RevisionMethod


def revise_forecasts(
    method: str,
    forecasts: Dict[str, numpy.ndarray],
    errors: Optional[Dict[str, numpy.ndarray]] = None,
    residuals: Optional[Dict[str, numpy.ndarray]] = None,
    summing_matrix: numpy.ndarray = None,
    nodes: NAryTreeT = None,
    transformer: TransformT = None,
):
    """
    Convenience function to get revised forecast for pre-computed base forecasts

    Parameters
    ----------
    method : str
        The reconciliation method to use
    forecasts : Dict[str, numpy.ndarray]
        A dict mapping key name to its forecasts (including in-sample forecasts). Required.
    errors : Dict[str, numpy.ndarray]
        A dict mapping key name to the in-sample errors. Required for methods: ``OLS``, ``WLSS``, ``WLSV``
    residuals : Dict[str, numpy.ndarray]
        A dict mapping key name to the residuals of in-sample forecasts. Required for methods: OLS, WLSS, WLSV
    summing_matrix : numpy.ndarray
        Not required if ``nodes`` argument is passed, or if using ``BU`` approach
    nodes : NAryTreeT
        The tree of nodes as specified in :py:class:`HierarchyTree <hts.hierarchy.HierarchyTree>`. Required if not
        if using ``AHP``, ``PHA` ``FP`` methods, or if using  passing the ``OLS``, ``WLSS``, ``WLSV`` methods
          and not passing the ``summing_matrix`` parameter
    transformer : TransformT
        A transform with the method: ``inv_func`` that will be applied to the forecasts

    Returns
    -------
    revised forecasts : pandas.DataFrame
        The revised forecasts
    """

    if nodes:
        summing_matrix, sum_mat_labels = to_sum_mat(nodes)

    if method in [MethodT.AHP.name, MethodT.PHA.name, MethodT.FP.name] and not nodes:
        raise ValueError(f"Method {method} requires an NAryTree to be passed")

    if method in [MethodT.OLS.name, MethodT.WLSS.name, MethodT.WLSV.name]:
        if not (all([forecasts, errors, residuals]) or (not summing_matrix)):
            raise ValueError(
                f"Method {method} requires forecasts, errors, and residuals to be passed, as "
                f"well as an NAryTree or a summing matrix"
            )

    revision = RevisionMethod(
        name=method, sum_mat=summing_matrix, transformer=transformer
    )

    revised = revision.revise(forecasts=forecasts, mse=errors, nodes=nodes)

    return pandas.DataFrame(revised, columns=list(forecasts.keys()))



from hts.core.regressor import HTSRegressor
from hts.revision import RevisionMethod

__author__ = """Carlo Mazzaferro"""
__email__ = "carlo.mazzaferro@gmail.com"
__version__ = "0.5.4"


__all__ = [
    "HTSRegressor",
    "RevisionMethod",
]



import pandas

from hts import HTSRegressor
from hts.core.result import HTSResult


def test_instantiate_regressor():
    ht = HTSRegressor()
    assert ht.model == "prophet"
    assert isinstance(ht.hts_result, HTSResult)
    assert isinstance(ht.hts_result.residuals, dict)


def test_fit_regressor_uv(load_df_and_hier_uv):
    hierarchical_sine_data, sine_hier = load_df_and_hier_uv
    ht = HTSRegressor(model="holt_winters", revision_method="OLS")
    ht.fit(df=hierarchical_sine_data, nodes=sine_hier)
    assert isinstance(ht.hts_result.errors, dict)
    assert isinstance(ht.hts_result.models, dict)
    assert "total" in ht.hts_result.models


def test_predict_regressor_uv(load_df_and_hier_uv):
    hierarchical_sine_data, sine_hier = load_df_and_hier_uv
    hsd = hierarchical_sine_data.head(200)

    for model in ["holt_winters", "auto_arima", "sarimax"]:
        ht = HTSRegressor(model=model, revision_method="OLS")

        ht.fit(df=hsd, nodes=sine_hier)
        preds = ht.predict(steps_ahead=10)

        assert isinstance(preds, pandas.DataFrame)

        # base + steps ahead
        assert len(preds) == len(hsd) + 10
        assert len(ht.hts_result.forecasts["total"]) == len(hsd) + 10
        assert "total" in ht.hts_result.errors
        assert "total" in ht.hts_result.forecasts
        assert "total" in ht.hts_result.residuals

        assert "a" in ht.hts_result.errors
        assert "a" in ht.hts_result.forecasts
        assert "a" in ht.hts_result.residuals


def test_fit_regressor_visnights(load_df_and_hier_visnights):
    hierarchical_vis_data, vis_hier = load_df_and_hier_visnights
    ht = HTSRegressor(model="holt_winters", revision_method="OLS")
    ht.fit(df=hierarchical_vis_data, nodes=vis_hier)
    assert isinstance(ht.hts_result.errors, dict)
    assert isinstance(ht.hts_result.models, dict)
    assert "total" in ht.hts_result.models


def test_predict_regressor_visnights(load_df_and_hier_visnights):
    hierarchical_vis_data, vis_hier = load_df_and_hier_visnights
    hvd = hierarchical_vis_data
    for model in ["holt_winters", "auto_arima"]:
        ht = HTSRegressor(model=model, revision_method="OLS")

        ht.fit(df=hvd, nodes=vis_hier)
        preds = ht.predict(steps_ahead=4)

        assert isinstance(preds, pandas.DataFrame)

        # base + steps ahead
        assert len(preds) == len(hvd) + 4
        assert len(ht.hts_result.forecasts["total"]) == len(hvd) + 4
        assert "total" in ht.hts_result.errors
        assert "total" in ht.hts_result.forecasts
        assert "total" in ht.hts_result.residuals

        assert "NSW" in ht.hts_result.errors
        assert "NSW" in ht.hts_result.forecasts
        assert "NSW" in ht.hts_result.residuals



import logging
from typing import Union

import numpy
import pandas
from scipy.special import inv_boxcox
from scipy.stats import boxcox
from sklearn.base import BaseEstimator, TransformerMixin

logger = logging.getLogger(__name__)


class FunctionTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, func: callable = boxcox, inv_func: callable = inv_boxcox):

        self.func = func
        self.inv_func = inv_func
        self.lam = None

    def fit(self, x: pandas.Series, y=None, **fit_params):
        return self

    def transform(self, x: pandas.Series):
        if any(x.values == 0):
            x += 1
            x, self.lam = self.func(x.values)
            return x
        elif any(x.values < 0):
            raise ValueError("Boxcox can\t be applied, column has negative values")
        else:
            x, self.lam = self.func(x.values)
            return x

    def fit_transform(self, x: pandas.Series, y=None, **fit_params):
        return self.fit(x).transform(x)

    def inverse_transform(self, x: Union[pandas.Series, numpy.ndarray]):
        if isinstance(x, pandas.Series):
            return self.inv_func(x.values, self.lam)
        else:
            return self.inv_func(x, self.lam)



import abc
import logging
import weakref
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    List,
    NamedTuple,
    NewType,
    Optional,
    Tuple,
    Union,
)
from weakref import ReferenceType

import numpy
import pandas
from sklearn.base import BaseEstimator, RegressorMixin

logger = logging.getLogger(__name__)


class ExtendedEnum(Enum):
    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))

    @classmethod
    def names(cls):
        return list(map(lambda c: c.name, cls))


class ModelT(str, ExtendedEnum):
    prophet = "prophet"
    holt_winters = "holt_winters"
    auto_arima = "auto_arima"
    sarimax = "sarimax"


class UnivariateModelT(str, ExtendedEnum):
    arima = "arima"
    auto_arima = "auto_arima"
    prophet = "prophet"
    holt_winters = "holt_winters"
    sarimax = "sarimax"


Models = NewType("ModelT", ModelT)


class Transform(NamedTuple):
    func: Callable
    inv_func: Callable


class HierarchyVisualizerT(metaclass=abc.ABCMeta):
    tree: "NAryTreeT"

    def create_map(self):
        ...


class NAryTreeT(metaclass=abc.ABCMeta):
    """
    Type definition of an NAryTree
    """

    key: str
    item: Union[pandas.Series, pandas.DataFrame]
    exogenous: List[str] = None
    children: List[Optional["NAryTreeT"]]
    _parent: "Optional[ReferenceType[NAryTreeT]]"
    visualizer: HierarchyVisualizerT

    @property
    def parent(self):
        if self._parent:
            return self._parent()

    def __iter__(self) -> "NAryTreeT":
        yield self
        for child in self.children:
            yield child

    def __getstate__(self):
        self._parent = None
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__ = state
        for child in self.children:
            child._parent = weakref.ref(self)

    def num_nodes(self) -> int:
        ...

    def is_leaf(self) -> bool:
        ...

    def value_at_height(self, level: int) -> List:
        ...

    def sum_at_height(self, level) -> int:
        ...

    def get_height(self) -> int:
        ...

    def level_order_traversal(self: "NAryTreeT") -> List[List[int]]:
        ...

    def get_level_order_labels(self: "NAryTreeT") -> List[List[str]]:
        ...

    def traversal_level(self) -> List["NAryTreeT"]:
        ...

    def add_child(self, key=None, item=None, exogenous=None) -> "NAryTreeT":
        ...

    def leaf_sum(self) -> int:
        ...

    def to_pandas(self) -> pandas.DataFrame:
        ...

    def get_node_height(self, key: str) -> int:
        ...

    def get_series(self) -> pandas.Series:
        ...

    def string_repr(self, prefix="", _last=True):
        base = "".join([prefix, "- " if _last else "|- ", self.key, "\n"])
        prefix += "   " if _last else "|  "
        child_count = len(self.children)
        for i, child in enumerate(self.children):
            _last = i == (child_count - 1)
            base += child.string_repr(_last=_last, prefix=prefix)
        return base

    def __repr__(self):
        return self.string_repr()

    __str__ = __repr__


class TimeSeriesModelT(BaseEstimator, RegressorMixin, metaclass=abc.ABCMeta):

    """ Type definition of an TimeSeriesModel """

    kind: str
    node: NAryTreeT
    model: Any
    forecast: Optional[pandas.DataFrame]
    residual: Optional[numpy.ndarray]
    mse: Optional[numpy.ndarray]
    transform: "TransformT"

    @staticmethod
    def _no_func(x):
        ...

    def _set_results_return_self(self, in_sample, y_hat) -> "TimeSeriesModelT":
        ...

    def create_model(self, **kwargs):
        ...

    def fit(self, **fit_args) -> "TimeSeriesModelT":
        raise NotImplementedError

    def predict(self, node: NAryTreeT, **predict_args):
        raise NotImplementedError


class MethodT(ExtendedEnum):
    OLS = "OLS"
    WLSS = "WLSS"
    WLSV = "WLSV"
    FP = "FP"
    PHA = "PHA"
    AHP = "AHP"
    BU = "BU"
    NONE = "NONE"


# TODO: make this a proper recursive type when mypy supports it: https://github.com/python/mypy/issues/731
HierarchyT = Tuple[str, "HierarchyT"]
NodesT = ExogT = Dict[str, List[str]]
LowMemoryFitResultT = Tuple[str, str]
ModelFitResultT = Union[TimeSeriesModelT, LowMemoryFitResultT]
HTSFitResultT = List[ModelFitResultT]
TransformT = Optional[Union[Transform, bool]]



# -*- coding: utf-8 -*-
# Many thanks to @blue-yonder for providing the base implementation for this file.
# see more at: https://github.com/blue-yonder/tsfresh

from itertools import chain

import numpy as np
import pytest
from distributed import Client

from hts import HTSRegressor
from hts.utilities.distribution import (
    ClusterDaskDistributor,
    LocalDaskDistributor,
    MultiprocessingDistributor,
)


def test_partition():
    distributor = MultiprocessingDistributor(n_workers=1)

    data = [1, 3, 10, -10, 343.0]
    distro = distributor.partition(data, 3)
    assert next(distro), [1, 3, 10]
    assert next(distro), [-10, 343.0]

    data = np.arange(10)
    distro = distributor.partition(data, 2)
    assert next(distro), [0, 1]
    assert next(distro), [2, 3]


@pytest.mark.serial
def test_calculate_best_chunk_size():
    distributor = MultiprocessingDistributor(n_workers=2)
    assert distributor.calculate_best_chunk_size(10), 1
    assert distributor.calculate_best_chunk_size(11), 2
    assert distributor.calculate_best_chunk_size(100), 10
    assert distributor.calculate_best_chunk_size(101), 11

    distributor = MultiprocessingDistributor(n_workers=3)
    assert distributor.calculate_best_chunk_size(10), 1
    assert distributor.calculate_best_chunk_size(30), 2
    assert distributor.calculate_best_chunk_size(31), 3


#


@pytest.mark.serial
def test_multiprocessing_fit(load_df_and_hier_uv):
    distributor = MultiprocessingDistributor(n_workers=2)
    hsd, hier = load_df_and_hier_uv
    reg = HTSRegressor()
    reg = reg.fit(df=hsd, nodes=hier, distributor=distributor)

    node_keys = list(chain.from_iterable([v for v in hier.values()]))

    for node in node_keys:
        assert node in reg.hts_result.models.keys()


@pytest.mark.serial
def test_dask_cluster_two_workers(load_df_and_hier_uv):
    with Client(n_workers=1, processes=False) as client:
        address = client.scheduler_info()["address"]
        distributor = ClusterDaskDistributor(address=address)

        hsd, hier = load_df_and_hier_uv
        reg = HTSRegressor()
        reg = reg.fit(df=hsd, nodes=hier, distributor=distributor)

        node_keys = list(chain.from_iterable([v for v in hier.values()]))

        for node in node_keys:
            assert node in reg.hts_result.models.keys()



import numpy
import pandas
import pytest

from hts.transforms import FunctionTransformer


def test_fit_transform(hierarchical_mv_data):
    tf = FunctionTransformer()
    df = hierarchical_mv_data
    transformed = tf.fit_transform(df["WF-01"])
    assert isinstance(transformed, numpy.ndarray)
    assert tf.lam is not None

    transformed = tf.fit_transform(df["WF-01"].replace(0, 6))
    assert isinstance(transformed, numpy.ndarray)
    assert tf.lam is not None

    with pytest.raises(ValueError):
        _ = tf.fit_transform(df["precipitation"].replace(0, -10))


def test_inv_transform(hierarchical_mv_data):
    tf = FunctionTransformer()
    df = hierarchical_mv_data

    base = df["WF-01"].replace(0, 6)
    transformed = tf.fit_transform(base)

    inv_transformed = tf.inverse_transform(transformed)
    assert numpy.allclose(inv_transformed, base, atol=0.1)

    inv_transformed = tf.inverse_transform(pandas.Series(transformed))
    assert numpy.allclose(inv_transformed, base, atol=0.1)



from typing import Dict, Optional, Tuple

import pandas


class HTSResult:
    def __init__(self) -> None:
        self._revised_forecasts: Optional[Dict] = None
        self._models: Dict = dict()
        self._errors: Dict = dict()
        self._residuals: Dict = dict()
        self._forecasts: Dict = dict()

    @property
    def forecasts(self) -> Dict:
        return self._forecasts

    @forecasts.setter
    def forecasts(self, kv_tuple: Tuple) -> None:
        k, v = kv_tuple
        self._forecasts[k] = v

    @property
    def errors(self) -> Dict:
        return self._errors

    @errors.setter
    def errors(self, kv_tuple: Tuple) -> None:
        k, v = kv_tuple
        self._errors[k] = v

    @property
    def residuals(self) -> Dict:
        return self._residuals

    @residuals.setter
    def residuals(self, kv_tuple: Tuple) -> None:
        k, v = kv_tuple
        self._residuals[k] = v

    @property
    def models(self) -> Dict:
        return self._models

    @models.setter
    def models(self, kv_tuple: Tuple) -> None:
        k, v = kv_tuple
        self._models[k] = v

    def to_pandas(self, p: str) -> pandas.DataFrame:
        kv = getattr(self, p)
        return pandas.DataFrame(kv)

    def get_series(self, p, key) -> pandas.Series:
        prop = getattr(self, p)
        return prop[key]



from hts._t import ModelT
from hts.hierarchy import HierarchyTree
from hts.model.base import TimeSeriesModel


class HoltWintersModel(TimeSeriesModel):
    """
    Wrapper class around ``statsmodels.tsa.holtwinters.ExponentialSmoothing``

    Attributes
    ----------
    model : ExponentialSmoothing
        The instance of the model

    _model : HoltWintersResults
        The result of model fitting. See statsmodels.tsa.holtwinters.HoltWintersResults

    mse : float
        MSE for in-sample predictions

    residual : numpy.ndarry
        Residuals for the in-sample predictions

    forecast : pandas.DataFramer
        The forecast for the trained model

    Methods
    -------
    fit(self, **fit_args)
        Fits underlying models to the data, passes kwargs to ``SARIMAX``

    predict(self, node, steps_ahead: int = 10)
        Predicts the n-step ahead forecast
    """

    def __init__(self, node: HierarchyTree, **kwargs):
        self._model = None
        super().__init__(ModelT.holt_winters.name, node, **kwargs)

    def predict(self, node: HierarchyTree, steps_ahead=10):
        y_hat = self._model.forecast(steps=steps_ahead).values
        in_sample_preds = self._model.predict(start=0, end=-1).values
        return self._set_results_return_self(in_sample_preds, y_hat)

    def fit(self, **fit_args) -> "TimeSeriesModel":
        self._model = self.model.fit(**fit_args)
        return self

    def fit_predict(self, node: HierarchyTree, steps_ahead=10, **fit_args):
        return self.fit(**fit_args).predict(node=node, steps_ahead=steps_ahead)



import os
from typing import Any


class suppress_stdout_stderr(object):
    """
    A context manager for doing a "deep suppression" of stdout and stderr in
    Python, i.e. will suppress all print, even if the print originates in a
    compiled C/Fortran sub-function.
       This will not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).

    """

    def __init__(self) -> None:
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self) -> None:
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

    def __exit__(self, *_: Any) -> None:
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fds[0], 1)
        os.dup2(self.save_fds[1], 2)
        # Close the null files
        for fd in self.null_fds + self.save_fds:
            os.close(fd)



from hts._t import ModelT
from hts.model.ar import AutoArimaModel, SarimaxModel
from hts.model.es import HoltWintersModel
from hts.model.p import FBProphetModel

__all__ = [
    "AutoArimaModel",
    "SarimaxModel",
    "HoltWintersModel",
    "FBProphetModel",
    "MODEL_MAPPING",
]


MODEL_MAPPING = {
    ModelT.auto_arima.name: AutoArimaModel,
    ModelT.holt_winters.name: HoltWintersModel,
    ModelT.prophet.name: FBProphetModel,
    ModelT.sarimax.name: SarimaxModel,
}



import numpy
import pandas
import pytest

from hts import HTSRegressor
from hts.convenience import revise_forecasts


def test_convenience_error_handling(load_df_and_hier_uv):
    hd, hier = load_df_and_hier_uv
    dummy = numpy.array([1, 2, 3, 5])
    with pytest.raises(ValueError):
        revise_forecasts("FP", forecasts={"a": dummy}, nodes=None)

    with pytest.raises(ValueError):
        revise_forecasts("OLS", summing_matrix=dummy, forecasts={"a": dummy})


def test_convenience(load_df_and_hier_uv):
    hd, hier = load_df_and_hier_uv
    hd = hd.head(200)
    ht = HTSRegressor(model="holt_winters", revision_method="OLS")
    ht = ht.fit(df=hd, nodes=hier)
    preds = ht.predict(steps_ahead=10)

    rev = revise_forecasts(
        "OLS",
        nodes=ht.nodes,
        forecasts=ht.hts_result.forecasts,
        errors=ht.hts_result.errors,
        residuals=ht.hts_result.residuals,
    )

    assert isinstance(rev, pandas.DataFrame)
    assert numpy.allclose(preds.values, rev.values)

    rev_s = revise_forecasts(
        "OLS",
        summing_matrix=ht.sum_mat,
        forecasts=ht.hts_result.forecasts,
        errors=ht.hts_result.errors,
        residuals=ht.hts_result.residuals,
    )
    assert isinstance(rev_s, pandas.DataFrame)
    assert numpy.allclose(rev.values, rev_s.values)

    rev = revise_forecasts("AHP", nodes=ht.nodes, forecasts=ht.hts_result.forecasts)
    assert isinstance(rev, pandas.DataFrame)



import os
import pickle
from typing import Dict, List, Optional, Tuple

import numpy
import pandas

from hts._t import (
    HTSFitResultT,
    LowMemoryFitResultT,
    ModelFitResultT,
    NAryTreeT,
    TimeSeriesModelT,
)
from hts.hierarchy.utils import make_iterable
from hts.utilities.distribution import (
    DistributorBaseClass,
    MapDistributor,
    MultiprocessingDistributor,
)


def _do_fit(
    nodes: NAryTreeT,
    function_kwargs,
    n_jobs: int,
    disable_progressbar: bool,
    show_warnings: bool,
    distributor: Optional[DistributorBaseClass],
) -> HTSFitResultT:

    distributor = _get_distributor(
        n_jobs=n_jobs,
        disable_progressbar=disable_progressbar,
        show_warnings=show_warnings,
        distributor=distributor,
    )

    result = distributor.map_reduce(
        _do_actual_fit, data=nodes, function_kwargs=function_kwargs
    )
    distributor.close()
    return result


def _do_actual_fit(node: NAryTreeT, function_kwargs: Dict) -> ModelFitResultT:
    instantiated_model = function_kwargs["model_instance"](
        node=node,
        transform=function_kwargs["transform"],
        **function_kwargs["model_args"]
    )
    if not function_kwargs["low_memory"]:
        model_instance = instantiated_model.fit(**function_kwargs["fit_kwargs"])
        return model_instance
    else:

        return _fit_serialize_model(instantiated_model, function_kwargs)


def _fit_serialize_model(
    model: TimeSeriesModelT, function_kwargs: Dict
) -> LowMemoryFitResultT:
    tmp = function_kwargs["tmp_dir"]
    path = os.path.join(tmp, model.node.key + ".pkl")
    model_instance = model.fit(**function_kwargs["fit_kwargs"])
    with open(path, "wb") as p:
        pickle.dump(model_instance, p)
    return model.node.key, path


def _do_predict(
    models: List[Tuple[str, ModelFitResultT, NAryTreeT]],
    function_kwargs: Dict,
    n_jobs: int,
    disable_progressbar: bool,
    show_warnings: bool,
    distributor: DistributorBaseClass,
) -> HTSFitResultT:

    distributor = _get_distributor(
        n_jobs=n_jobs,
        disable_progressbar=disable_progressbar,
        show_warnings=show_warnings,
        distributor=distributor,
    )

    result = distributor.map_reduce(
        _do_actual_predict, data=models, function_kwargs=function_kwargs
    )
    distributor.close()
    return result


def _model_mapping_to_iterable(
    model_mapping: Dict[str, ModelFitResultT], nodes: NAryTreeT
) -> List[Tuple[str, ModelFitResultT, NAryTreeT]]:
    prediction_triplet = []

    for node in make_iterable(nodes, prop=None):
        if isinstance(model_mapping[node.key], tuple):
            model = model_mapping[node.key][1]
        else:
            model = model_mapping[node.key]
        prediction_triplet.append((node.key, model, node))
    return prediction_triplet


def _do_actual_predict(
    model: Tuple[str, ModelFitResultT, NAryTreeT], function_kwargs: Dict
) -> Tuple[str, pandas.DataFrame, numpy.ndarray, numpy.ndarray]:
    key, file_or_model, node = model
    if function_kwargs["low_memory"]:
        model_instance = _load_serialized_model(
            tmp_dir=function_kwargs["tmp_dir"], file_name=file_or_model
        )
    else:
        model_instance = file_or_model
    model_instance = model_instance.predict(
        node=node,
        steps_ahead=function_kwargs["steps_ahead"],
        **function_kwargs["predict_kwargs"]
    )
    return key, model_instance.forecast, model_instance.mse, model_instance.residual


def _load_serialized_model(tmp_dir, file_name):
    path = os.path.join(tmp_dir, file_name)
    with open(path, "rb") as p:
        return pickle.load(p)


def _get_distributor(
    n_jobs: int,
    disable_progressbar: bool,
    show_warnings: bool,
    distributor: Optional[DistributorBaseClass],
):
    if distributor is None:
        if n_jobs == 0:
            distributor = MapDistributor(
                disable_progressbar=disable_progressbar,
                progressbar_title="Fitting models: ",
            )
        else:
            distributor = MultiprocessingDistributor(
                n_workers=n_jobs,
                disable_progressbar=disable_progressbar,
                progressbar_title="Fitting models",
                show_warnings=show_warnings,
            )

    if not isinstance(distributor, DistributorBaseClass):
        raise ValueError("the passed distributor is not an DistributorBaseClass object")
    return distributor



from random import choice
from typing import Dict, List, Tuple

import numpy as np
import pandas

from hts._t import MethodT, NAryTreeT
from hts.hierarchy import make_iterable


def to_sum_mat(ntree: NAryTreeT) -> Tuple[np.ndarray, List[str]]:
    """
    This function creates a summing matrix for the bottom up and optimal combination approaches
    All the inputs are the same as above
    The output is a summing matrix, see Rob Hyndman's "Forecasting: principles and practice" Section 9.4

    Parameters
    ----------
    ntree : NAryTreeT


    Returns
    -------
    numpy.ndarray
        Summing matrix.

    List[str]
        Row order list of the level in the hierarchy represented by each row in the summing matrix.

    """
    nodes = ntree.level_order_traversal()
    node_labels = ntree.get_level_order_labels()
    num_at_level = list(map(sum, nodes))
    columns = num_at_level[-1]

    # Initialize summing matrix with bottom level rows
    sum_mat = np.identity(columns)

    # Names of each row in summing matrix.
    sum_mat_labels = []

    # Bottom level matrix labels, with indices correspoding to column in summing matrix
    bl_mat_idx_ref = node_labels[-1]

    # Skip total and bottom level of tree. Rows added outside of loop.
    for level in node_labels[1:-1]:
        for label in level:
            # Exclude duplicates specified in tree
            if label not in sum_mat_labels:
                row = []
                for bl_element in bl_mat_idx_ref:
                    # Check if the bottom level element is part of label
                    is_component = all(
                        [True if x in bl_element else False for x in label.split("_")]
                    )
                    if is_component:
                        row.append(1)
                    else:
                        row.append(0)

                # Add row correspoding to label to top of summing matrix
                row = np.array(row)
                sum_mat = np.vstack((row, sum_mat))
                sum_mat_labels.append(label)

    # Add top as first row in summing matrix
    top = np.ones(columns)
    sum_mat = np.vstack((top, sum_mat))

    # Reverse list of labels to match summing matrix, since vstack and append worked in the opposite order.
    # Not currently returned, but could be for information or matrix alignment.
    sum_mat_labels.reverse()
    sum_mat_labels = ["total"] + sum_mat_labels + bl_mat_idx_ref

    return sum_mat, sum_mat_labels


def project(
    hat_mat: np.ndarray, sum_mat: np.ndarray, optimal_mat: np.ndarray
) -> np.ndarray:
    new_mat = np.empty([hat_mat.shape[0], sum_mat.shape[0]])
    for i in range(hat_mat.shape[0]):
        new_mat[i, :] = np.dot(optimal_mat, np.transpose(hat_mat[i, :]))
    return new_mat


def y_hat_matrix(forecasts, keys=None):
    if not keys:
        keys = forecasts.keys()
    first = list(forecasts.keys())[0]
    y_hat_mat = np.zeros([len(forecasts[first].yhat), 1])
    for key in keys:
        f1 = np.array(forecasts[key].yhat)
        f2 = f1[:, np.newaxis]
        if np.all(y_hat_mat == 0):
            y_hat_mat = f2
        else:
            y_hat_mat = np.concatenate((y_hat_mat, f2), axis=1)
    return y_hat_mat


def optimal_combination(
    forecasts: Dict[str, pandas.DataFrame],
    sum_mat: np.ndarray,
    method: str,
    mse: Dict[str, float],
):
    """
    Produces the optimal combination of forecasts by trace minimization (as described by
    Wickramasuriya, Athanasopoulos, Hyndman in "Optimal Forecast Reconciliation for Hierarchical and Grouped Time
    Series Through Trace Minimization")

    Parameters
    ----------
    forecasts : dict
        Dictionary of pandas.DataFrames containing the future predictions
    sum_mat : np.ndarray
        The summing  matrix
    method : str
        One of:
            - OLS (ordinary least squares)
            - WLSS (structurally weighted least squares)
            - WLSV (variance weighted least squares)
    mse

    Returns
    -------

    """
    hat_mat = y_hat_matrix(forecasts)
    transpose = np.transpose(sum_mat)

    if method == MethodT.OLS.name:
        ols = np.dot(
            np.dot(sum_mat, np.linalg.inv(np.dot(transpose, sum_mat))), transpose
        )
        return project(hat_mat=hat_mat, sum_mat=sum_mat, optimal_mat=ols)
    elif method == MethodT.WLSS.name:
        diag = np.diag(np.transpose(np.sum(sum_mat, axis=1)))
    elif method == MethodT.WLSV.name:
        diag = [mse[key] for key in mse.keys()]
        diag = np.diag(np.flip(np.hstack(diag) + 0.0000001, 0))
    else:
        raise ValueError("Invalid method")

    # S*inv(S'S)*S'
    optimal_mat = np.dot(
        np.dot(
            np.dot(
                sum_mat,
                np.linalg.inv(np.dot(np.dot(transpose, np.linalg.inv(diag)), sum_mat)),
            ),
            transpose,
        ),
        np.linalg.inv(diag),
    )

    return project(hat_mat=hat_mat, sum_mat=sum_mat, optimal_mat=optimal_mat)


def proportions(nodes, forecasts, sum_mat, method=MethodT.PHA.name):
    n_cols = len(list(forecasts.keys()))
    fcst = forecasts[list(forecasts.keys())[0]].yhat
    fcst = fcst[:, np.newaxis]
    num_bts = sum_mat.shape[1]

    cols = [n.key for n in [nodes] + nodes.traversal_level()][
        (n_cols - num_bts) : n_cols
    ]

    bts_dat = nodes.to_pandas()[cols]
    if method == MethodT.AHP.name:
        divs = np.divide(np.transpose(np.array(bts_dat)), np.array(nodes.get_series()))
        props = divs.mean(1)
        props = props[:, np.newaxis]
    elif method == MethodT.PHA.name:
        bts_sum = bts_dat.sum(0)
        top_sum = sum(nodes.get_series())
        props = bts_sum / top_sum
        props = props[:, np.newaxis]
    else:
        raise ValueError("Invalid method")
    return np.dot(np.array(fcst), np.transpose(props))


def forecast_proportions(forecasts, nodes):
    """
    Cons:
       Produces biased revised forecasts even if base forecasts are unbiased
    """
    n_cols = len(list(forecasts.keys())) + 1

    levels = nodes.get_height()
    column = 0
    first_node = 1

    key = choice(list(forecasts.keys()))
    new_mat = np.empty([len(forecasts[key].yhat), n_cols - 1])
    new_mat[:, 0] = forecasts[key].yhat

    as_iterable = make_iterable(nodes, prop=None)

    for level in range(levels - 1):
        for i, node in enumerate(nodes.level_order_traversal()[level]):
            num_child = node
            last_node = first_node + num_child
            base_fcst = np.array(
                [forecasts[k.key].yhat[:] for k in as_iterable[first_node:last_node]]
            )
            print(base_fcst.shape)
            fore_sum = np.sum(base_fcst, axis=0)
            fore_sum = fore_sum[:, np.newaxis]
            if column == 0:
                rev_top = np.array(forecasts["total"].yhat)
                rev_top = rev_top[:, np.newaxis]
            else:
                rev_top = np.array(new_mat[:, column])
                rev_top = rev_top[:, np.newaxis]
            new_mat[:, first_node:last_node] = np.divide(
                np.multiply(np.transpose(base_fcst), rev_top), fore_sum
            )
            column += 1
            first_node += num_child
    return new_mat



import logging

import numpy
import pandas

from hts._t import ModelT
from hts.hierarchy import HierarchyTree
from hts.model.base import TimeSeriesModel
from hts.utilities.utils import suppress_stdout_stderr

logger = logging.getLogger(__name__)

logging.getLogger("fbprophet").setLevel(logging.ERROR)


class FBProphetModel(TimeSeriesModel):
    """
    Wrapper class around ``fbprophet.Prophet``

    Attributes
    ----------
    model : Prophet
        The instance of the model

    mse : float
        MSE for in-sample predictions

    residual : numpy.ndarry
        Residuals for the in-sample predictions

    forecast : pandas.DataFramer
        The forecast for the trained model

    Methods
    -------
    fit(self, **fit_args)
        Fits underlying models to the data, passes kwargs to ``fbprophet.Prophet``

    predict(self, node, steps_ahead: int = 10, freq: str = 'D', **predict_args)
        Predicts the n-step ahead forecast. Exogenous variables are required if models were
        fit using them, frequency should be passed as well
    """

    def __init__(self, node: HierarchyTree, **kwargs):
        super().__init__(ModelT.prophet.name, node, **kwargs)
        self.cap = None
        self.floor = None
        self.include_history = False

    def create_model(self, capacity_max=None, capacity_min=None, **kwargs):
        self.cap = capacity_max
        self.floor = capacity_min

        if not capacity_max and not capacity_min:
            growth = "linear"
        else:
            growth = "logistic"
            if self.cap:
                self.node.item["cap"] = capacity_max
            if self.floor:
                self.node.item["floor"] = capacity_min

        try:
            from fbprophet import Prophet
        except ImportError:  # pragma: no cover
            logger.error(
                "prophet model requires fbprophet to work. Exiting."
                "Install it with: pip install scikit-hts[prophet]"
            )
            return
        model = Prophet(growth=growth, **kwargs)
        if self.node.exogenous:
            for ex in self.node.exogenous:
                model.add_regressor(ex)
        return model

    def _reformat(self, node):
        if isinstance(node, pandas.Series):
            node = pandas.DataFrame(node)
        df = node.rename(columns={self.node.key: "y"})
        df["ds"] = pandas.to_datetime(df.index)
        return df.reset_index(drop=True)

    def fit(self, **fit_args) -> "TimeSeriesModel":
        df = self._reformat(self.node.item)
        with suppress_stdout_stderr():
            self.model = self.model.fit(df)
            self.model.stan_backend = None
        return self

    def predict(self, node: HierarchyTree, freq: str = "D", steps_ahead: int = 1):

        df = self._reformat(node.item)
        future = self.model.make_future_dataframe(
            periods=steps_ahead, freq=freq, include_history=True
        )
        if self.cap:
            future["cap"] = self.cap
        if self.floor:
            future["floor"] = self.floor

        self.forecast = self.model.predict(future)
        merged = pandas.merge(df, self.forecast, on="ds")
        self.residual = (merged["yhat"] - merged["y"]).values
        self.mse = numpy.mean(numpy.array(self.residual) ** 2)
        if self.cap is not None:
            self.forecast.yhat = numpy.exp(self.forecast.yhat)
        if self.transform:
            self.forecast.yhat = self.transformer.inverse_transform(self.forecast.yhat)
            self.forecast.trend = self.transformer.inverse_transform(
                self.forecast.trend
            )
            for component in ["seasonal", "daily", "weekly", "yearly", "holidays"]:
                if component in self.forecast.columns.tolist():
                    inv_transf = self.transformer.inverse_transform(
                        getattr(self.forecast, component)
                    )
                    setattr(self.forecast, component, inv_transf)
        return self



import logging
import tempfile
from typing import Any, Dict, List, Optional, Union

import numpy
import pandas
from sklearn.base import BaseEstimator, RegressorMixin

from hts import defaults
from hts import model as hts_models
from hts._t import ExogT, MethodT, ModelT, NodesT, TimeSeriesModelT, Transform
from hts.core.exceptions import InvalidArgumentException, MissingRegressorException
from hts.core.result import HTSResult
from hts.core.utils import _do_fit, _do_predict, _model_mapping_to_iterable
from hts.functions import to_sum_mat
from hts.hierarchy import HierarchyTree
from hts.hierarchy.utils import make_iterable
from hts.model.base import TimeSeriesModel
from hts.revision import RevisionMethod
from hts.utilities.distribution import DistributorBaseClass

logger = logging.getLogger(__name__)


class HTSRegressor(BaseEstimator, RegressorMixin):
    """
    Main regressor class for scikit-hts. Likely the only import you'll need for using
    this project. It takes a pandas dataframe, the nodes specifying the hierarchies, model kind, revision
    method, and a few other parameters. See Examples to get an idea of how to use it.


    Attributes
    ----------
    transform : Union[NamedTuple[str, Callable], bool]
        Function transform to be applied to input and outputs. If True, it will use ``scipy.stats.boxcox``
         and ``scipy.special._ufuncs.inv_boxcox`` on input and output data

    sum_mat : array_like
        The summing matrix, explained in depth in `Forecasting <https://otexts.com/fpp2/gts.html>`_

    nodes : Dict[str, List[str]]
        Nodes representing node, edges of the hierarchy. Keys are nodes, values are list of edges.

    df : pandas.DataFrame
        The dataframe containing the nodes and edges specified above

    revision_method : str
        One of: ``"OLS", "WLSS", "WLSV", "FP", "PHA", "AHP", "BU", "NONE"``

    models : dict
        Dictionary that holds the trained models

    mse : dict
        Dictionary that holds the mse scores for the trained models

    residuals : dict
        Dictionary that holds the mse residual for the trained models

    forecasts : dict
        Dictionary that holds the forecasts for the trained models

    model_instance : TimeSeriesModel
        Reference to the class implementing the actual time series model

    """

    def __init__(
        self,
        model: str = defaults.MODEL,
        revision_method: str = defaults.REVISION,
        transform: Optional[Union[Transform, bool]] = None,
        n_jobs: int = defaults.N_PROCESSES,
        low_memory: bool = defaults.LOW_MEMORY,
        **kwargs: Any,
    ):
        """
        Parameters
        ----------
        model : str
            One of the models supported by ``hts``. These can be found
        revision_method : str
            The revision method to be used. One of: ``"OLS", "WLSS", "WLSV", "FP", "PHA", "AHP", "BU", "NONE"``
        transform : Boolean or NamedTuple
            If True, ``scipy.stats.boxcox`` and ``scipy.special._ufuncs.inv_boxcox`` will be applied prior and after
            fitting.
            If False, no transform is applied.
            If you desired to use custom functions, use a NamedTuple like: ``{'func': Callable, 'inv_func': Callable}``
        n_jobs : int
            Number of parallel jobs to run the forecasting on
        low_memory : Bool
            If True, models will be fit, serialized, and released from memory. Usually a good idea if
            you are dealing with a large amount of nodes
        kwargs
            Keyword arguments to be passed to the underlying model to be instantiated
        """

        self.model = model
        self.method: str = revision_method
        self.n_jobs: int = n_jobs
        self.low_memory: bool = low_memory
        if self.low_memory:
            self.tmp_dir: Optional[str] = tempfile.mkdtemp(prefix="hts_")
        else:
            self.tmp_dir = None
        self.transform = transform

        self.sum_mat: Optional[numpy.ndarray] = None
        self.nodes: Optional[NodesT] = None
        self.model_instance: Optional[TimeSeriesModelT] = None
        self.exogenous: bool = False
        self.revision_method: Optional[RevisionMethod] = None
        self.hts_result: HTSResult = HTSResult()
        self.model_args = kwargs

    def __init_hts(
        self,
        nodes: Optional[NodesT] = None,
        df: Optional[pandas.DataFrame] = None,
        tree: Optional[HierarchyTree] = None,
        root: str = "root",
        exogenous: Optional[List[str]] = None,
    ):

        if not nodes and not df:
            if not tree:
                raise InvalidArgumentException(
                    "Either nodes and df must be passed, or a pre-built hierarchy tree"
                )
            else:
                self.nodes = tree
        else:
            self.nodes = HierarchyTree.from_nodes(
                nodes=nodes, df=df, exogenous=exogenous, root=root
            )
        self.exogenous = exogenous
        self.sum_mat, sum_mat_labels = to_sum_mat(self.nodes)
        self._set_model_instance()
        self._init_revision()

    def _init_revision(self):
        self.revision_method = RevisionMethod(
            sum_mat=self.sum_mat, transformer=self.transform, name=self.method
        )

    def _set_model_instance(self):
        try:
            self.model_instance = hts_models.MODEL_MAPPING[self.model]
        except KeyError:
            raise InvalidArgumentException(
                f'Model {self.model} not valid. Pick one of: {" ".join(ModelT.names())}'
            )

    def fit(
        self,
        df: Optional[pandas.DataFrame] = None,
        nodes: Optional[NodesT] = None,
        tree: Optional[HierarchyTree] = None,
        exogenous: Optional[ExogT] = None,
        root: str = "total",
        distributor: Optional[DistributorBaseClass] = None,
        disable_progressbar=defaults.DISABLE_PROGRESSBAR,
        show_warnings=defaults.SHOW_WARNINGS,
        **fit_kwargs: Any,
    ) -> "HTSRegressor":

        """
        Fit hierarchical model to dataframe containing hierarchical data as specified in the ``nodes`` parameter
        Exogenous can also be passed as a dict of (string, list), where string is the specific node key and the list
        contains the names of the columns to be used as exogenous variables for that node.

        Alternatively, a pre-built HierarchyTree can be passed without specifying the node and df. See more at
        :class:`hts.hierarchy.HierarchyTree`

        Parameters
        ----------
        df : pandas.DataFrame
            A Dataframe of time series with a DateTimeIndex. Each column represents a node in the hierarchy. Ignored if
            tree argument is passed
        nodes : Dict[str, List[str]]
            The hierarchy defined as a dict of (string, list), as specified in
             :py:func:`HierarchyTree.from_nodes <hts.hierarchy.HierarchyTree.from_nodes>`
        tree : HierarchyTree
            A pre-built HierarchyTree. Ignored if df and nodes are passed, as the tree will be built from thise
        distributor : Optional[DistributorBaseClass]
             A distributor, for parallel/distributed processing
        exogenous : Dict[str, List[str]] or None
            Node key mapping to columns that contain the exogenous variable for that node
        root : str
            The name of the root node
        disable_progressbar : Bool
            Disable or enable progressbar
        show_warnings : Bool
            Disable warnings
        fit_kwargs : Any
            Any arguments to be passed to the underlying forecasting model's fit function

        Returns
        -------
        HTSRegressor
            The fitted HTSRegressor instance
        """

        self.__init_hts(nodes=nodes, df=df, tree=tree, root=root, exogenous=exogenous)

        nodes = make_iterable(self.nodes, prop=None)

        fit_function_kwargs = {
            "fit_kwargs": fit_kwargs,
            "low_memory": self.low_memory,
            "tmp_dir": self.tmp_dir,
            "model_instance": self.model_instance,
            "model_args": self.model_args,
            "transform": self.transform,
        }

        fitted_models = _do_fit(
            nodes=nodes,
            function_kwargs=fit_function_kwargs,
            n_jobs=self.n_jobs,
            disable_progressbar=disable_progressbar,
            show_warnings=show_warnings,
            distributor=distributor,
        )

        for model in fitted_models:
            if isinstance(model, tuple):
                self.hts_result.models = model
            else:
                self.hts_result.models = (model.node.key, model)
        return self

    def __init_predict_step(self, exogenous_df: pandas.DataFrame, steps_ahead: int):
        if self.exogenous and not exogenous_df:
            raise MissingRegressorException(
                "Exogenous variables were provided at fit step, hence are required at "
                "predict step. Please pass the 'exogenous_df' variable to predict "
                "function"
            )
        if not exogenous_df and not steps_ahead:
            logger.info(
                "No arguments passed for 'steps_ahead', defaulting to predicting 1-step-ahead"
            )
            steps_ahead = 1
        elif exogenous_df:
            steps_ahead = len(exogenous_df)
            for node in make_iterable(self.nodes, prop=None):
                exog_cols = node.exogenous
                try:
                    node.item = exogenous_df[exog_cols]
                except KeyError:
                    raise MissingRegressorException(
                        f"Node {node.key} has as exogenous variables {node.exogenous} but "
                        f"these columns were not found in 'exogenous_df'"
                    )
        return steps_ahead

    def predict(
        self,
        exogenous_df: pandas.DataFrame = None,
        steps_ahead: int = None,
        distributor: Optional[DistributorBaseClass] = None,
        disable_progressbar: bool = defaults.DISABLE_PROGRESSBAR,
        show_warnings: bool = defaults.SHOW_WARNINGS,
        **predict_kwargs,
    ) -> pandas.DataFrame:
        """

        Parameters
        ----------
        distributor : Optional[DistributorBaseClass]
             A distributor, for parallel/distributed processing
        disable_progressbar : Bool
            Disable or enable progressbar
        show_warnings : Bool
            Disable warnings
        predict_kwargs : Any
            Any arguments to be passed to the underlying forecasting model's predict function
        exogenous_df : pandas.DataFrame
            A dataframe of lenght == steps_ahead containing the exogenous data for each of the nodes
        steps_ahead : int
            The number of forecasting steps for which to produce a forecast

        Returns
        -------
        Revised Forescasts, as a pandas.DataFrame in the same format as the one passed for fitting, extended by `steps_ahead`
        time steps`
        """

        steps_ahead = self.__init_predict_step(exogenous_df, steps_ahead)
        predict_function_kwargs = {
            "fit_kwargs": predict_kwargs,
            "steps_ahead": steps_ahead,
            "low_memory": self.low_memory,
            "tmp_dir": self.tmp_dir,
            "predict_kwargs": predict_kwargs,
        }

        fit_models = _model_mapping_to_iterable(self.hts_result.models, self.nodes)
        results = _do_predict(
            models=fit_models,
            function_kwargs=predict_function_kwargs,
            n_jobs=self.n_jobs,
            disable_progressbar=disable_progressbar,
            show_warnings=show_warnings,
            distributor=distributor,
        )
        for key, forecast, error, residual in results:
            self.hts_result.forecasts = (key, forecast)
            self.hts_result.errors = (key, error)
            self.hts_result.residuals = (key, residual)
        return self._revise(steps_ahead=steps_ahead)

    def _revise(self, steps_ahead: int = 1) -> pandas.DataFrame:
        logger.info(f"Reconciling forecasts using {self.revision_method}")
        revised = self.revision_method.revise(
            forecasts=self.hts_result.forecasts,
            mse=self.hts_result.errors,
            nodes=self.nodes,
        )

        revised_columns = list(make_iterable(self.nodes))
        revised_index = self._get_predict_index(steps_ahead=steps_ahead)
        return pandas.DataFrame(revised, index=revised_index, columns=revised_columns)

    def _get_predict_index(self, steps_ahead=1) -> Any:

        freq = getattr(self.nodes.item.index, "freq", 1)
        start = self.nodes.item.index[-1] + freq
        end = self.nodes.item.index[-1] + (steps_ahead * freq)

        future = pandas.date_range(freq=freq, start=start, end=end)

        return self.nodes.item.index.append(future)



import logging
import string
from itertools import chain

import numpy

from hts._t import HierarchyVisualizerT, NAryTreeT

logger = logging.getLogger(__name__)


def get_min_max_ll(geos):
    fl = list(chain.from_iterable([g[0] for g in geos]))
    mx_lat = max([x[0] for x in fl])
    mx_lon = max([x[1] for x in fl])

    mn_lat = min([x[0] for x in fl])
    mn_lon = min([x[1] for x in fl])

    return mx_lat, mx_lon, mn_lat, mn_lon


class HierarchyVisualizer(HierarchyVisualizerT):
    def __init__(self, tree: NAryTreeT):
        self.tree = tree

    @property
    def as_df(self):
        return self.tree.to_pandas()

    def get_geos(self):
        try:
            from h3 import h3
        except ImportError:  # pragma: no cover
            logger.error(
                "h3-py must be installed for geo hashing capabilities. Exiting."
                "Install it with: pip install scikit-hts[geo]"
            )
            return
        h3s = [
            col for col in self.as_df.columns if all(c in string.hexdigits for c in col)
        ]
        return [
            (h3.h3_to_geo_boundary(g), self.as_df[g].fillna(0).sum(), g) for g in h3s
        ]

    def h3_to_lat_long(self):
        return

    def create_map(self):

        try:
            import branca.colormap as cm
            from folium import Map
            from folium.vector_layers import Polygon
        except ImportError:  # pragma: no cover
            logger.error(
                "Mapping requires folium==0.10.0 to be installed, geo mapping will not work."
                "Install it with: pip install scikit-hts[geo]"
            )
            return

        _map = Map(tiles="cartodbpositron")
        geos = self.get_geos()
        max_lat, max_lon, min_lat, min_lon = get_min_max_ll(geos)

        geos = [
            (i, numpy.log(j + 1) / (self.tree.get_node_height(k) + 1), k)
            for i, j, k in geos
        ]
        mx, mn = max([j for i, j, k in geos]), min([j for i, j, k in geos])
        geos = [(i, (j - mn) / (mx - mn), k) for i, j, k in geos]

        for points, count, h in sorted(geos, key=lambda x: x[1]):
            tooltip = f"hex: {h}"
            polygon = Polygon(
                locations=points,
                tooltip=tooltip,
                fill=True,
                color=cm.linear.OrRd_03.rgb_hex_str(count),
                fill_color=cm.linear.OrRd_03.rgb_hex_str(count),
                fill_opacity=0.3,
                weight=3,
                opacity=0.4,
            )
            polygon.add_to(_map)

        _map.fit_bounds([[min_lat, min_lon], [max_lat, max_lon]])

        return _map



import numpy

from hts import HTSRegressor
from hts.revision import RevisionMethod


def test_instantiate_revision(load_df_and_hier_visnights):
    hierarchical_visnights_data, visnights_hier = load_df_and_hier_visnights

    for method in ["OLS", "FP", "WLSS", "WLSV", "PHA", "AHP", "NONE"]:
        ht = HTSRegressor(model="holt_winters", revision_method=method)
        ht.fit(df=hierarchical_visnights_data, nodes=visnights_hier)

        rm = RevisionMethod(method, sum_mat=ht.sum_mat, transformer=ht.transform)

        _ = ht.predict(steps_ahead=3)
        revised = rm.revise(
            forecasts=ht.hts_result.forecasts, mse=ht.hts_result.errors, nodes=ht.nodes
        )
        assert isinstance(revised, numpy.ndarray)
        assert revised.shape == (11, len(ht.hts_result.forecasts))
